% How to use this class:
%  1. define a moviePlayer object with M = moviePlayer;
%  2. set the object's properties for movieData_xyt (an array sized nx * ny * nt),
%     xgridmat, ygridmat, tvec. Set the xlabel_text, ylabel_text, zlabel_text, and
%     zlabel_unit_str proeprties, and define a function for the tstringfunc property
%     (should take 1 input argument, which is the frame number -- a value from 1 to nt)
%  3. call the makeMovieFigure method with 'M.makeMovieFigure'
%  4. 'play' the movie with the playButtonPressed method with 'M.playButtonPressed'

% TODO: add hidden line removal toggle
classdef moviePlayer < handle
    properties
        movieData_xyt       (:,:,:) double {mustBeReal}                                 = zeros(51,51,101);
        xgridmat            (:,:) double {mustBeReal, mustBeFinite}                     = repmat(linspace(-1,1,51),[51,1]);
        ygridmat            (:,:) double {mustBeReal, mustBeFinite}                     = repmat(linspace(-1,1,51).',[1,51]);
        tvec                (:,1) double {mustBeReal, mustBeFinite}                     = linspace(-1,1,101);
        % movieFrame          (1,1) double {mustBePositive, mustBeReal, mustBeFinite}     = 1;
        figtag              (1,1) string = "movieFig";
        fighandle           (1,1)
        axhandle            (1,1) 
        meshhandle          (1,1)
        titlehandle         (1,1)
        frameslider         (1,1)
        frameeditbox        (1,1)
        framecontrolheight  (1,1) double {mustBeReal,mustBePositive,mustBeFinite} = 22;
        framecontrolboxwidth(1,1) double {mustBeReal,mustBePositive,mustBeFinite} = 50;
        norm_movie          (1,1) logical = false;
        buttonwidth         (1,1) double {mustBeReal,mustBePositive,mustBeFinite} = 45;
        xlabel_text         (1,1) string = "x [mm]";
        ylabel_text         (1,1) string = "y [mm]";
        zlabel_text         (1,1) string = "Irradiance";
        zlabel_unit_str     (1,1) string = "[W/m^2]";
        pause_value         (1,1) double {mustBeNonnegative} = 0.05; % 'pause' this number of seconds between movie frames
        tstringfunc % Should be function handle that takes 1 input argument (frame number) and outputs a string
    end
    properties (SetAccess = private)
        playbutton          (1,1)
        resetbutton         (1,1)
        norm_checkbox       (1,1)
        hlr_checkbox        (1,1)
    end

    methods
        function out = moviePlayer(varargin)
            % Constructor class -- don't expect any input arguments; if there are any,
            % keyboard
            if nargin ~= 0
                keyboard;
            end
            out.tstringfunc = @(J) sprintf('%g s', out.tvec(J));
            if nargout == 0
                clear('out');
            end
        end

        % function set.movieFrame(obj,val)
        %     arguments
        %         obj moviePlayer;
        %         val uint16 {mustBePositive, mustBeReal, mustBeFinite};
        %     end
        %     try 
        %         setFrame(obj, val);
        %         set(obj.frameslider, 'value', val);
        %         set(obj.frameeditbox, 'string', sprintf('%i',val));
        %     catch E
        %         keyboard;
        %     end
        % end

        
        function fh = makeMovieFigure(obj)
            % Use the figure obj.fighandle or make a new figure and assign it to obj.fighandle
            if isempty(obj.fighandle) || ~ishandle(obj.fighandle) || (obj.fighandle == 0)
                fh = findobj('tag', obj.figtag);
                if isempty(fh)
                    fh = figure('tag', obj.figtag);
                    obj.fighandle = fh;
                else
                    obj.fighandle = fh;
                end
            end

            % Clear figure
            clf(obj.fighandle);

            % xgridmat and ygridmat properties of moviePlayer class must be same size (and are meshgrids)
            if ~isequal(size(obj.xgridmat), size(obj.ygridmat))
                error('moviePlayer property xgridmat must be same shape as ygridmat. Instead, xgridmat is %ix%i and ygridmat is %ix%i.', ...
                    size(obj.xgridmat,1), size(obj.xgridmat,2), size(obj.ygridmat,1), size(obj.ygridmat,2));
            end

            % movieData_xyt array size in 1st & 2nd dimensions must be equal to xgridmat size (and ygridmat)
            if ~ ( isequal(size(obj.xgridmat,1), size(obj.movieData_xyt,1)) && isequal(size(obj.xgridmat,2), size(obj.movieData_xyt,2)) )
                error('The first two dimensions of the moviePlayer property movieData_xyt must have same shape as xgridmat & ygridmat. Instead, movieData_xyt is %ix%ix%i and ygridmat is %ix%i.', ...
                    size(obj.movieData_xyt,1), size(obj.movieData_xyt,2), size(obj.movieData_xyt,3), size(obj.ygridmat,1), size(obj.ygridmat,2));
            end
            % Make new axes object in figure
            obj.axhandle    = axes('parent', obj.fighandle);
            % Make new wiremesh plot in axes we just created
            obj.meshhandle  = mesh(obj.axhandle, obj.xgridmat, obj.ygridmat, obj.movieData_xyt(:,:,1));
            % Turn on 'box' for axes, calculate x, y, and z limits
            box(obj.axhandle, 'on');
            xl = [min(obj.xgridmat, [], 'all'), max(obj.xgridmat, [], 'all')];
            yl = [min(obj.ygridmat, [], 'all'), max(obj.ygridmat, [], 'all')];
            zl = [min(obj.movieData_xyt, [], 'all'), max(obj.movieData_xyt, [], 'all')];
            if diff(zl) == 0
                zl = [0,1e-12];
            end
            zlim(obj.axhandle, zl);
            xlim(obj.axhandle, xl);
            ylim(obj.axhandle, yl);

            daspect(obj.axhandle, [diff(xl), diff(yl), diff(zl)]);

            % Make title of axes
            if (~isempty(obj.tstringfunc)) && (isa(obj.tstringfunc, 'function_handle'))
                % obj.titlehandle = title(obj.axhandle, obj.tstringfunc(obj.movieFrame), 'fontweight', 'normal');
                obj.titlehandle = title(obj.axhandle, obj.tstringfunc(1), 'fontweight', 'normal');
            else
                obj.titlehandle = title(obj.axhandle, '', 'fontweight', 'normal');
            end
            
            % When the figure with window is resized, have to update positions of axes and
            % movie control strip (progress bar, frame number edit box, play & reset
            % pushbuttons, and the normalize checkbox)
            set(obj.fighandle, 'ResizeFcn', @obj.resizeMovie);
            
            % 
            moviestrip_h    = obj.framecontrolheight; % Height in pixels of the movie control strip at the bottom of the figure -- it'll have the slider, and editbox with frame index
            framebox_w      = obj.framecontrolboxwidth;
            y0 = 3; % y start for movie control strip at bottom of figure
            set(obj.fighandle, 'units', 'pixels');
            ip = get(obj.fighandle, 'position');
            set(obj.axhandle, 'units', 'pixels', 'outerposition',[0,moviestrip_h+2+y0,ip(3),ip(4)-moviestrip_h-2]);
            obj.playbutton = uicontrol('parent', obj.fighandle, 'style', 'pushbutton', 'string', 'Play', 'Position',[5,y0,obj.buttonwidth,obj.framecontrolheight], ...
                'Callback', @obj.playButtonPressed);
            obj.resetbutton = uicontrol('parent', obj.fighandle, 'style', 'pushbutton', 'string', 'Reset', 'Position',[5+obj.buttonwidth,y0,obj.buttonwidth,obj.framecontrolheight], ...
                'Callback', @obj.resetButtonPressed, 'Enable', 'off');

            obj.norm_checkbox = uicontrol('parent', obj.fighandle, 'style', 'checkbox', 'string', 'Normalized', 'Value', 0, 'Position', [5, y0 + obj.framecontrolheight, 85, obj.framecontrolheight]);
            obj.hlr_checkbox = uicontrol('parent', obj.fighandle, 'style', 'checkbox', 'string', 'HLR', 'Value', 1, 'Position', [5, y0 + obj.framecontrolheight*2, 85, obj.framecontrolheight], 'Tooltip', 'Hidden Line Removal (opaque wire mesh)');

            obj.frameslider = uicontrol('parent', obj.fighandle, 'Style', 'slider', 'units', 'pixels', 'value', 1, ...
                'Min', 1, 'Max', length(obj.tvec), 'Position', [5+2*obj.buttonwidth, y0, ip(3)-framebox_w-5-2*obj.buttonwidth, moviestrip_h], ...
                'SliderStep', [1,5]./(length(obj.tvec)-1), ...
                'callback', @obj.frameSliderCallback);
            obj.frameeditbox = uicontrol('parent', obj.fighandle, 'style', 'edit', 'string', '1', 'units', 'pixels', ...
                'Position', [ip(3)-framebox_w, y0 ,framebox_w, moviestrip_h], 'callback', @obj.frameBoxCallback);

            if ~verLessThan('matlab','9.5')
                addToolbarExplorationButtons(obj.fighandle);
            end
            if ~isempty(obj.xlabel_text)
                xlabel(obj.axhandle, obj.xlabel_text);
            end
            if ~isempty(obj.ylabel_text)
                ylabel(obj.axhandle, obj.ylabel_text);
            end
            if (~isempty(obj.zlabel_text)) && (~isempty(obj.zlabel_unit_str))
                zlabel(obj.axhandle, obj.zlabel_text + obj.zlabel_unit_str);
            end

            % daspect can be tricky; if you have unequal x & y ranges should they be shown
            % as square? or should a larger x range look bigger?
            
            % Just make them square:
            % daspect(obj.axhandle, [diff(xl),diff(yl),diff(zl)]);

            % Try to make axes proportional: use daspect(1,1,x), but what to use for x?
            % keyboard;
        end
        function resizeMovie(obj, figh, evt)
            y0 = 3;
            % keyboard;
            oldunits = get(obj.fighandle, 'units');
            set(obj.fighandle, 'units', 'pixels');
            drawnow;
            ip = get(obj.fighandle, 'Position');
            set(obj.axhandle, 'units', 'pixels', 'outerposition', [0, obj.framecontrolheight, ip(3), ip(4)-obj.framecontrolheight-y0-3]);
            set(obj.frameeditbox, 'units', 'pixels', 'position', [ip(3)-obj.framecontrolboxwidth-3, y0, obj.framecontrolboxwidth ,obj.framecontrolheight]);
            % set(obj.frameslider, 'position', [5, y0, ip(3)-obj.framecontrolboxwidth-8, obj.framecontrolheight]);
            set(obj.frameslider,   'position', [5+2*obj.buttonwidth, y0, ip(3)-obj.framecontrolboxwidth-5-2*obj.buttonwidth, obj.framecontrolheight]);
            set(obj.fighandle,     'units', oldunits);
            set(obj.playbutton,    'position', [5,y0,obj.buttonwidth,obj.framecontrolheight]);
            set(obj.resetbutton,   'position', [5+obj.buttonwidth,y0,obj.buttonwidth,obj.framecontrolheight]);
            set(obj.norm_checkbox, 'position', [5, y0 + obj.framecontrolheight, 85, obj.framecontrolheight]);
            set(obj.hlr_checkbox, 'position', [5, y0 + obj.framecontrolheight*2, 85, obj.framecontrolheight]);
        end
        % function ax = makeMovieAxis(obj)
        %     keyboard;
        % end
        function varargout = frameBoxCallback(obj, thebox, evt, varargin)
            % Callback for when the string in the 'frame' edit box changes
            if nargin ~= 3
                keyboard;
            end
            frame = fix(str2double(get(obj.frameeditbox, 'string')));
            maxframe = length(obj.tvec);
            if frame < 1
                frame = 1;
            elseif frame > maxframe
                frame = maxframe;
            end
            set(obj.frameslider, 'value', frame);
            obj.setFrame(frame);
        end
        function varargout = frameSliderCallback(obj, sliderobj, evt, varargin)
            % Callback for when the frame slider is used
            % return;
            if nargin ~= 3
                keyboard;
            end
            val = fix(get(sliderobj, 'value'));
            set(obj.frameeditbox, 'string', sprintf('%i', val));
            obj.setFrame(val);
            % keyboard;
        end
        function varargout = playButtonPressed(obj, playbutton, evt, varargin)
            % Callback for when the 'Play' button has been pressed
            if (nargin ~= 3) && (nargin ~= 1)
                keyboard;
            end
            playbutton = obj.playbutton;
            buttonstr = get(playbutton,'string');
            if strcmpi(buttonstr, 'Play')
                set(playbutton,'string','Stop');
            elseif strcmpi(buttonstr, 'Stop')
                set(playbutton,'string','Play');
                return
            end
            set(obj.resetbutton,'enable','off');
            sliderFrame = get(obj.frameslider,'value');
            boxFrame = str2double(get(obj.frameeditbox,'string'));
            if sliderFrame ~= boxFrame
                currentFrame = fix(( sliderFrame + boxFrame) / 2);
            else
                currentFrame = sliderFrame;
            end
            if currentFrame < 1
                currentFrame = 1;
            elseif currentFrame >= length(obj.tvec)
                currentFrame = 1;
            end
            for frame = currentFrame : length(obj.tvec)
                buttonstr = get(playbutton,'string');
                if strcmpi(buttonstr, 'Play')
                    set(obj.resetbutton,'enable','on');
                    return
                end
                obj.setFrame(frame);
                set(obj.frameslider,'value',frame);
                set(obj.frameeditbox,'string',sprintf('%i',frame));
                % pause(0.05);
                pause(obj.pause_value);
            end
            set(playbutton,'string','Play');
            set(obj.resetbutton,'enable','on');
        end
        function varargout = resetButtonPressed(obj, resetbutton, evt)
            % Callback for when the 'Reset' button has been pressed
            % keyboard;
            obj.setFrame(1);
            set(obj.frameslider,'value', 1);
            set(obj.frameeditbox,'string', '1');
            set(resetbutton, 'enable', 'off');
        end
        function varargout = setFrame(obj, framenumber)
            % Callback for when 
            normalized = get(obj.norm_checkbox, 'value');
            hlr = get(obj.hlr_checkbox, 'value');
            if ~normalized
                set(obj.meshhandle, 'zdata', obj.movieData_xyt(:,:,framenumber));
                if (~isempty(obj.tstringfunc)) && (isa(obj.tstringfunc, 'function_handle'))
                    set(obj.titlehandle, 'string', obj.tstringfunc(framenumber));
                end
                % zl = [0, max(obj.movieData_xyt,[],'all')];
                zl = [min(0, min(obj.movieData_xyt,[],'all')), max(obj.movieData_xyt,[],'all')];
                if diff(zl) == 0
                    zl = [0,1e-12];
                end
                zlim(obj.axhandle, zl);
                xl = xlim(obj.axhandle);
                yl = ylim(obj.axhandle);
                % daspect(obj.axhandle, [diff(xl), diff(yl), 1]);
                daspect(obj.axhandle, [diff(xl), diff(yl), diff(zl)]);
                if (~isempty(obj.zlabel_text)) && (~isempty(obj.zlabel_unit_str))
                    zlabel(obj.axhandle, obj.zlabel_text + " " + obj.zlabel_unit_str);
                end
                drawnow;
            else
                set(obj.meshhandle, 'zdata', obj.movieData_xyt(:,:,framenumber) ./ max(obj.movieData_xyt(:,:,framenumber), [], 'all'));
                if (~isempty(obj.tstringfunc)) && (isa(obj.tstringfunc, 'function_handle'))
                    set(obj.titlehandle, 'string', obj.tstringfunc(framenumber));
                end
                zlim(obj.axhandle, [0, 1]);
                xl = xlim(obj.axhandle);
                yl = ylim(obj.axhandle);
                daspect(obj.axhandle, [diff(xl), diff(yl), 1]);
                if (~isempty(obj.zlabel_text))
                    zlabel(obj.axhandle, "Normalized " + obj.zlabel_text + " [arb]");
                end
                drawnow;
            end
            if hlr
                % set(obj.meshhandle, 'facecolor', 'white');
                set(obj.meshhandle, 'facecolor', get(obj.axhandle,'color'));
            else
                set(obj.meshhandle, 'facecolor', 'none');
            end
        end
    end % end methods
end % end classdef